/* Tree-dumping functionality for intermediate representation.
   Copyright (C) 1999, 2000, 2002, 2003, 2004, 2005
   Free Software Foundation, Inc.
   Written by Mark Mitchell <mark@codesourcery.com>

Modifications copyright 2001,2002 by James Michael DuPont <mdupont777@yahoo.com>
This file is part of GCC.

GCC is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

GCC is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#include "config.h"
#include "system.h"
#include "coretypes.h"
#include "tm.h"
#include "tree.h"
#include "splay-tree.h"
#include "diagnostic.h"
#include "toplev.h"
#include "tree-dump.h"
#include "tree-pass.h"
#include "langhooks.h"
#include "tree-iterator.h"

#include "introspector.h"
#include "tree-dump-data.h"

unsigned int queue (dump_info_p, tree, int);
void dump_index (dump_info_p, unsigned int);

void dump_new_line (dump_info_p);
void dump_maybe_newline (dump_info_p);

int dump_enable_all (int, int);

/* Add T to the end of the queue of nodes to dump.  Returns the index
   assigned to T.  */

unsigned int
queue (dump_info_p di, tree t, int flags)
{
  dump_queue_p dq;
  dump_node_info_p dni;
  unsigned int index;

  /* Assign the next available index to T.  */
  index = ++di->index;

  /* Obtain a new queue node.  */
  if (di->free_list)
    {
      dq = di->free_list;
      di->free_list = dq->next;
    }
  else
    dq = xmalloc (sizeof (struct dump_queue));

  /* Create a new entry in the splay-tree.  */
  dni = xmalloc (sizeof (struct dump_node_info));
  dni->index = index;
  dni->binfo_p = ((flags & DUMP_BINFO) != 0);
  dq->node = splay_tree_insert (di->nodes, (splay_tree_key) t,
				(splay_tree_value) dni);

  /* Add it to the end of the queue.  */
  dq->next = 0;
  if (!di->queue_end)
    di->queue = dq;
  else
    di->queue_end->next = dq;
  di->queue_end = dq;

  /* Return the index.  */
  return index;
}

void
dump_index (dump_info_p di, unsigned int index)
{
  index=0;
//    assert(di->stream);
  //CLEANprintf (di->stream, "Dumping @%-6u\n", index);
  //CLEAN
  //  fprintf (di->stream, "ID @%-6u ", index);
  di->column += 8;
}

/* If T has not already been output, queue it for subsequent output.
   FIELD is a string to print before printing the index.  Then, the
   index of T is printed.  */

void
queue_and_dump_index (introspector_dump_info_p idi, introspector_predicate_p field, tree t, int flags)
{
  dump_info_p di=idi->di;
  //  introspector_dump_info_p idi
  unsigned int index;

  splay_tree_node n;

  /* If there's no node, just return.  This makes for fewer checks in
     our callers.  */
  if (!t)
    return;

  /* See if we've already queued or dumped this node.  */
  n = splay_tree_lookup (di->nodes, (splay_tree_key) t);
  if (n)
    index = ((dump_node_info_p) n->value)->index;
  else
    /* If we haven't, add it to the queue.  */
    index = queue (di, t, flags);



  /* introspector :
    look up the field, dump a ref to another object, and also emit information about the type of that object
    the next step will be put put the different types of nodes into subdirs
   */
  //  ppredicate = intrspctr_get_field_predicate(idi,field);
  intrspctr_node_ref(idi,field,index);
  //  intrspctr_set_node_type(idi,ppredicate,t);
  // end of introspector
}

/* Dump the type of T.  */

void
queue_and_dump_type (introspector_dump_info_p idi, tree t)
{
  queue_and_dump_index (idi,INTROSPECTOR_FIELD(type), TREE_TYPE (t), DUMP_NONE);
}

/* Dump column control */
#define SOL_COLUMN 25		/* Start of line column.  */
#define EOL_COLUMN 55		/* End of line column.  */
#define COLUMN_ALIGNMENT 15	/* Alignment.  */

/* Insert a new line in the dump output, and indent to an appropriate
   place to start printing more fields.  */

void
dump_new_line (dump_info_p di)
{
  ////CLEAN  fprintf (di->stream, "\n%*s", SOL_COLUMN, "");
  //  fprintf (di->stream, "\n");
  di->column = SOL_COLUMN;
}

/* If necessary, insert a new line.  */

void
dump_maybe_newline (dump_info_p di)
{
  int extra;

  /* See if we need a new line.  */
  if (di->column > EOL_COLUMN)
    dump_new_line (di);
  /* See if we need any padding.  */
  else if ((extra = (di->column - SOL_COLUMN) % COLUMN_ALIGNMENT) != 0)
    {
      ////CLEAN      fprintf (di->stream, "%*s", COLUMN_ALIGNMENT - extra, "");
      //      fprintf (di->stream, "\n");
      di->column += COLUMN_ALIGNMENT - extra;
    }
}

/* Dump pointer PTR using FIELD to identify it.  */

void
dump_pointer (introspector_dump_info_p idi,introspector_predicate_p field, void *ptr)
{
  introspector_dump_pointer (idi, field,  ptr);
}

/* Dump integer I using FIELD to identify it.  */

void
dump_int (introspector_dump_info_p idi, introspector_predicate_p field, int i)
{
  introspector_dump_int (idi, field,  i);
}

/* Dump the string S.  */

void
dump_string (introspector_dump_info_p idi, const char *string)
{
  introspector_dump_string (idi, string);
}

/* 
   Dump the string field S
   the type of the predicate is the field name in the gcc model.
*/

void
dump_string_field (introspector_dump_info_p idi, introspector_predicate_p field, const char *string)
{
  // introspector replaces this function totally
  intrspctr_add_statement_literal_string(idi,

					 /* lookup this field  type*/
					 field,
					 /* dump this string */
					 string             
					   );
}

/* Dump the next node in the queue.  */

     //		  int follow_chains //// introspector adds this function
void
dequeue_and_dump (introspector_dump_info_p idi)
{
  dump_info_p di= idi->di;

  dump_queue_p dq;
  splay_tree_node stn;
  dump_node_info_p dni;
  tree t;
  unsigned int index;
  enum tree_code code;
  enum tree_code_class code_class;

//  assert(di->stream);
  /* Get the next node from the queue.  */
  dq = di->queue;
  stn = dq->node;
  t = (tree) stn->key;
  dni = (dump_node_info_p) stn->value;
  index = dni->index;

  /* Remove the node from the queue, and put it on the free list.  */
  di->queue = dq->next;
  if (!di->queue)
    di->queue_end = 0;
  dq->next = di->free_list;
  di->free_list = dq;

  dump_index_rdf_main(idi,index); // set th 

  /* Print the node index.  */
  //dump_index (di, index); // was dump_index, renamed by introspector

  /* And the type of node this is.  */

  intrspctr_set_field_treecode(idi,t);

  /* Figure out what kind of node this is.  */
  code = TREE_CODE (t);
  code_class = TREE_CODE_CLASS (code);

  /* Although BINFOs are TREE_VECs, we dump them specially so as to be
     more informative.  */
  if (dni->binfo_p)
    {
      // this is code that traverses the tree, n
      unsigned ix;
      tree base;
      VEC (tree,gc) *accesses = BINFO_BASE_ACCESSES (t);

      dump_child (INTROSPECTOR_FIELD(type), BINFO_TYPE (t));

      if (BINFO_VIRTUAL_P (t))
	dump_string (idi, "virt");

      dump_int (idi, INTROSPECTOR_FIELD(bases), BINFO_N_BASE_BINFOS (t));
      for (ix = 0; BINFO_BASE_ITERATE (t, ix, base); ix++)
	{
	  tree access = (accesses ? VEC_index (tree, accesses, ix)
			 : access_public_node);
	  const char *string = NULL;

	  if (access == access_public_node)
	    string = "pub";
	  else if (access == access_protected_node)
	    string = "prot";
	  else if (access == access_private_node)
	    string = "priv";
	  else
	    gcc_unreachable ();

	  dump_string (idi, string);
	  queue_and_dump_index_introspector (idi,INTROSPECTOR_FIELD(binf), base, DUMP_BINFO);
	}

      goto done;
    }

  /* We can knock off a bunch of expression nodes in exactly the same
     way.  */
  if (IS_EXPR_CODE_CLASS (code_class))
    {
      /* If we're dumping children, dump them now.  */
      queue_and_dump_type (idi, t);

      switch (code_class)
	{
	case tcc_unary:
	  dump_child(INTROSPECTOR_FIELD(op_0), TREE_OPERAND (t, 0));
	  break;

	case tcc_binary:
	case tcc_comparison:
	  dump_child(INTROSPECTOR_FIELD(op_0), TREE_OPERAND (t, 0));
	  dump_child(INTROSPECTOR_FIELD(op_1), TREE_OPERAND (t, 1));
	  break;

	case tcc_expression:
	case tcc_reference:
	case tcc_statement:
	  /* These nodes are handled explicitly below.  */
	  break;

	default:
	  gcc_unreachable ();
	}
    }
  else if (DECL_P (t))
    {
      expanded_location xloc;
      /* All declarations have names.  */
      if (DECL_NAME (t))
	dump_child(INTROSPECTOR_FIELD(name), DECL_NAME (t));
      if (DECL_ASSEMBLER_NAME_SET_P (t)
	  && DECL_ASSEMBLER_NAME (t) != DECL_NAME (t))
	dump_child(INTROSPECTOR_FIELD(mngl), DECL_ASSEMBLER_NAME (t));
      /* And types.  */
      queue_and_dump_type (idi, t);
      dump_child(INTROSPECTOR_FIELD(scpe), DECL_CONTEXT (t));
      /* And a source position.  */
      xloc = expand_location (DECL_SOURCE_LOCATION (t));
      if (xloc.file)
	{
	  const char *filename = strrchr (xloc.file, '/');
	  if (!filename)
	    filename = xloc.file;
	  else
	    /* Skip the slash.  */
	    ++filename;
	  // INTROSPECTOR:
	    if (strcmp(filename, "<built-in>") ==0)
	    { 
	      intrspctr_dump_filename(idi,"built-in",0); // line number is not needed
	    }
	    else	      
	    {
	      intrspctr_dump_filename(idi,filename,DECL_SOURCE_LINE (t)); // line number is not needed
	    }	    
	  // INTROSPECTOR
	}
      /* And any declaration can be compiler-generated.  */
      if (DECL_ARTIFICIAL (t))
	dump_string (idi, "artificial");

      // INTROSPECTOR
      /*
      if (follow_chains) // to avoid decl chaining inside of function bodies
	{
	  // The following code is the same
	  if (TREE_CHAIN (t) && !dump_flag (idi, TDF_SLIM, NULL))
	    dump_child(INTROSPECTOR_FIELD(chan), TREE_CHAIN (t));
	}
      */
      // INTROSPECTOR
    }
  else if (code_class == tcc_type)
    {
      /* All types have qualifiers.  */
      int quals = lang_hooks.tree_dump.type_quals (t);

      if (quals != TYPE_UNQUALIFIED)
	{
	  // introspector
	  // boolean fields
	  if (quals & TYPE_QUAL_CONST)
	    {
	      intrspctr_assert_field_true(idi,INTROSPECTOR_FIELD(const));
	    }

	  if (quals & TYPE_QUAL_VOLATILE)
	    {
	      intrspctr_assert_field_true(idi,INTROSPECTOR_FIELD(volitile));
	    }

	  if (quals & TYPE_QUAL_RESTRICT)
	    {
	      intrspctr_assert_field_true(idi,INTROSPECTOR_FIELD(restrict));
	    }
	  // introspector
	}

      /* All types have associated declarations.  */
      dump_child(INTROSPECTOR_FIELD(name), TYPE_NAME (t));

      /* All types have a main variant.  */
      if (TYPE_MAIN_VARIANT (t) != t)
	dump_child(INTROSPECTOR_FIELD(unql), TYPE_MAIN_VARIANT (t));

      /* And sizes.  */
      dump_child(INTROSPECTOR_FIELD(size), TYPE_SIZE (t));

      /* All types have alignments.  */
      dump_int (idi, INTROSPECTOR_FIELD(algn), TYPE_ALIGN (t));
    }
  else if (code_class == tcc_constant)
    /* All constants can have types.  */
    queue_and_dump_type (idi, t);

  /* Give the language-specific code a chance to print something.  If
     it's completely taken care of things, don't bother printing
     anything more ourselves.  */
  if (lang_hooks.tree_dump.dump_tree (idi, t))
    goto done;

  /* Now handle the various kinds of nodes.  */
  switch (code)
    {
      int i;

    case IDENTIFIER_NODE:
      dump_string_field (idi, INTROSPECTOR_FIELD(strg), IDENTIFIER_POINTER (t));
      //      dump_int (idi, INTROSPECTOR_FIELD(lngt), IDENTIFIER_LENGTH (t));
      break;

    case TREE_LIST:
      dump_child(INTROSPECTOR_FIELD(purp), TREE_PURPOSE (t));
      dump_child(INTROSPECTOR_FIELD(valu), TREE_VALUE (t));
      dump_child(INTROSPECTOR_FIELD(chan), TREE_CHAIN (t));  // TODO: maybe we need to turn this off on follow chains?
      break;

    case STATEMENT_LIST:
      {
	tree_stmt_iterator it;
	for (i = 0, it = tsi_start (t); !tsi_end_p (it); tsi_next (&it), i++)
	  {
	    //	    char buffer[64];
	    //sprintf (buffer, "child[%u]", i);
	    dump_child (INTROSPECTOR_FIELD(child_statement), tsi_stmt (it));
	  }
      }
      break;

    case TREE_VEC:
      dump_int (idi, INTROSPECTOR_FIELD(lngt), TREE_VEC_LENGTH (t));
      for (i = 0; i < TREE_VEC_LENGTH (t); ++i)
	{
	  //char buffer[64];
	  //	  sprintf (buffer, INTROSPECTOR_FIELD(vec_child), i);
	  dump_child (INTROSPECTOR_FIELD(vec_child), TREE_VEC_ELT (t, i));
	}
      break;

    case INTEGER_TYPE:
    case ENUMERAL_TYPE:
      dump_int (idi, INTROSPECTOR_FIELD(prec), TYPE_PRECISION (t));
      if (TYPE_UNSIGNED (t))
	dump_string (idi,"unsigned");
      dump_child(INTROSPECTOR_FIELD(min), TYPE_MIN_VALUE (t));
      dump_child(INTROSPECTOR_FIELD(max), TYPE_MAX_VALUE (t));

      if (code == ENUMERAL_TYPE)
	dump_child(INTROSPECTOR_FIELD(csts), TYPE_VALUES (t));
      break;

    case REAL_TYPE:
      dump_int (idi, INTROSPECTOR_FIELD(prec), TYPE_PRECISION (t));
      break;

    case POINTER_TYPE:
      dump_child(INTROSPECTOR_FIELD(ptd), TREE_TYPE (t));
      break;

    case REFERENCE_TYPE:
      dump_child(INTROSPECTOR_FIELD(refd), TREE_TYPE (t));
      break;

    case METHOD_TYPE:
      dump_child(INTROSPECTOR_FIELD(clas), TYPE_METHOD_BASETYPE (t));
      /* Fall through.  */

    case FUNCTION_TYPE:
      dump_child(INTROSPECTOR_FIELD(retn), TREE_TYPE (t));
      dump_child(INTROSPECTOR_FIELD(prms), TYPE_ARG_TYPES (t));
      break;

    case ARRAY_TYPE:
      dump_child(INTROSPECTOR_FIELD(elts), TREE_TYPE (t));
      dump_child(INTROSPECTOR_FIELD(domn), TYPE_DOMAIN (t));
      break;

    case RECORD_TYPE:
    case UNION_TYPE:
      if (TREE_CODE (t) == RECORD_TYPE)
	dump_string (idi,"struct");
      else
	dump_string (idi,"union");

      dump_child(INTROSPECTOR_FIELD(flds), TYPE_FIELDS (t));
      dump_child(INTROSPECTOR_FIELD(fncs), TYPE_METHODS (t));
      queue_and_dump_index (idi, INTROSPECTOR_FIELD(binf), TYPE_BINFO (t),
			    DUMP_BINFO);
      break;

    case CONST_DECL:
      dump_child(INTROSPECTOR_FIELD(cnst), DECL_INITIAL (t));
      break;

    case VAR_DECL:
    case PARM_DECL:
    case FIELD_DECL:
    case RESULT_DECL:
      if (TREE_CODE (t) == PARM_DECL)
	{
	  dump_child(INTROSPECTOR_FIELD(chan), TREE_CHAIN (t));
	  dump_child(INTROSPECTOR_FIELD(argt), DECL_ARG_TYPE (t));
	}
      else
	dump_child(INTROSPECTOR_FIELD(init), DECL_INITIAL (t));
      dump_child(INTROSPECTOR_FIELD(size), DECL_SIZE (t));
      dump_int (idi, INTROSPECTOR_FIELD(algn), DECL_ALIGN (t));


      if (TREE_CODE (t) == FIELD_DECL)
	{
	  dump_child(INTROSPECTOR_FIELD(chan), TREE_CHAIN (t));
	  if (DECL_FIELD_OFFSET (t))
	    dump_child(INTROSPECTOR_FIELD(bpos), bit_position (t));
	}
      else if (TREE_CODE (t) == VAR_DECL
	       || TREE_CODE (t) == PARM_DECL)
	{
	  dump_int (idi, INTROSPECTOR_FIELD(used), TREE_USED (t));
	  if (DECL_REGISTER (t))
	    dump_string (idi,"register");
	}

      if (TREE_CODE (t) == VAR_DECL)
	{
	  // added a child object to the dump for collecting all the vars.
	  dump_child(INTROSPECTOR_FIELD(chan), TREE_CHAIN (t));
	}

      break;

    case FUNCTION_DECL:
      dump_child(INTROSPECTOR_FIELD(args), DECL_ARGUMENTS (t));
      if (DECL_EXTERNAL (t))
	dump_string (idi,"undefined");
      if (TREE_PUBLIC (t))
	dump_string (idi,"extern");
      else
	dump_string (idi,"static");

//      if (DECL_LANG_SPECIFIC (t) && !dump_flag (idi, TDF_SLIM, t))
	dump_child(INTROSPECTOR_FIELD(body), DECL_SAVED_TREE (t));
      break;

    case INTEGER_CST:
      if (TREE_INT_CST_HIGH (t))
	dump_int (idi, INTROSPECTOR_FIELD(high), TREE_INT_CST_HIGH (t));
      dump_int (idi, INTROSPECTOR_FIELD(low), TREE_INT_CST_LOW (t));
      break;

    case STRING_CST:

      // Introspector:TODO, replace with a string literal!
      intrspctr_add_statement_field_string(
					   idi,
					   TREE_STRING_POINTER (t) 
					   );
            
      //      dump_int (idi, INTROSPECTOR_FIELD(lngt), TREE_STRING_LENGTH (t));
      break;

    case TRUTH_NOT_EXPR:
    case ADDR_EXPR:
    case INDIRECT_REF:
    case ALIGN_INDIRECT_REF:
    case MISALIGNED_INDIRECT_REF:
    case CLEANUP_POINT_EXPR:
    case SAVE_EXPR:
    case REALPART_EXPR:
    case IMAGPART_EXPR:
      /* These nodes are unary, but do not have code class `1'.  */
      dump_child(INTROSPECTOR_FIELD(op_0), TREE_OPERAND (t, 0));
      break;

    case TRUTH_ANDIF_EXPR:
    case TRUTH_ORIF_EXPR:
    case INIT_EXPR:
    case MODIFY_EXPR:
    case COMPOUND_EXPR:
    case PREDECREMENT_EXPR:
    case PREINCREMENT_EXPR:
    case POSTDECREMENT_EXPR:
    case POSTINCREMENT_EXPR:
      /* These nodes are binary, but do not have code class `2'.  */
      dump_child(INTROSPECTOR_FIELD(op_0), TREE_OPERAND (t, 0));
      dump_child(INTROSPECTOR_FIELD(op_1), TREE_OPERAND (t, 1));
      break;

    case COMPONENT_REF:
      dump_child(INTROSPECTOR_FIELD(op_0), TREE_OPERAND (t, 0));
      dump_child(INTROSPECTOR_FIELD(op_1), TREE_OPERAND (t, 1));
      dump_child(INTROSPECTOR_FIELD(op_2), TREE_OPERAND (t, 2));
      break;

    case ARRAY_REF:
    case ARRAY_RANGE_REF:
      dump_child(INTROSPECTOR_FIELD(op_0), TREE_OPERAND (t, 0));
      dump_child(INTROSPECTOR_FIELD(op_1), TREE_OPERAND (t, 1));
      dump_child(INTROSPECTOR_FIELD(op_2), TREE_OPERAND (t, 2));
      dump_child(INTROSPECTOR_FIELD(op_3), TREE_OPERAND (t, 3));
      break;

    case COND_EXPR:
      dump_child(INTROSPECTOR_FIELD(op_0), TREE_OPERAND (t, 0));
      dump_child(INTROSPECTOR_FIELD(op_1), TREE_OPERAND (t, 1));
      dump_child(INTROSPECTOR_FIELD(op_2), TREE_OPERAND (t, 2));
      break;

    case CALL_EXPR:
      dump_child(INTROSPECTOR_FIELD(fn), TREE_OPERAND (t, 0));
      dump_child(INTROSPECTOR_FIELD(args), TREE_OPERAND (t, 1));
      break;

    case CONSTRUCTOR:
      dump_child(INTROSPECTOR_FIELD(elts), CONSTRUCTOR_ELTS (t));
      break;

    case BIND_EXPR:
      dump_child(INTROSPECTOR_FIELD(vars), TREE_OPERAND (t, 0)); // the variables

      dump_child(INTROSPECTOR_FIELD(body), TREE_OPERAND (t, 1));
      break;

    case LOOP_EXPR:
      dump_child(INTROSPECTOR_FIELD(body), TREE_OPERAND (t, 0));
      break;

    case EXIT_EXPR:
      dump_child(INTROSPECTOR_FIELD(cond), TREE_OPERAND (t, 0));
      break;

    case RETURN_EXPR:
      dump_child(INTROSPECTOR_FIELD(retn), TREE_OPERAND (t, 0));
      break;

    case TARGET_EXPR:
      dump_child(INTROSPECTOR_FIELD(decl), TREE_OPERAND (t, 0));
      dump_child(INTROSPECTOR_FIELD(init), TREE_OPERAND (t, 1));
      dump_child(INTROSPECTOR_FIELD(clnp), TREE_OPERAND (t, 2));
      /* There really are two possible places the initializer can be.
	 After RTL expansion, the second operand is moved to the
	 position of the fourth operand, and the second operand
	 becomes NULL.  */
      dump_child(INTROSPECTOR_FIELD(init), TREE_OPERAND (t, 3));
      break;

    default:
      /* There are no additional fields to print.  */
      break;
    }

 done:
  //
  //break;
  return;
//  if (dump_flag (idi, TDF_ADDRESS, NULL))
//    dump_pointer (idi,INTROSPECTOR_FIELD(addr), (void *)t);
  /* Terminate the line.  */
  //CLEAN
  //    fprintf (di->stream, "\n");
}

/* Return nonzero if FLAG has been specified for the dump, and NODE
   is not the root node of the dump.  */

int dump_flag (dump_info_p di, int flag, tree node)
{
  return (di->flags & flag) && (node != di->node);
}


/* Table of tree dump switches. This must be consistent with the
   TREE_DUMP_INDEX enumeration in tree.h , moved out to tree-dump-data.c */

/* Dynamically registered tree dump files and switches.  */
static struct dump_file_info *extra_dump_files;
static size_t extra_dump_files_in_use;
static size_t extra_dump_files_alloced;

/* Define a name->number mapping for a dump flag value.  */
struct dump_option_value_info
{
  const char *const name;	/* the name of the value */
  const int value;		/* the value of the name */
};

/* Table of dump options. This must be consistent with the TDF_* flags
   in tree.h */
static const struct dump_option_value_info dump_options[] =
{
  {"address", TDF_ADDRESS},
  {"slimo", TDF_SLIM},
  {"raw", TDF_RAW},
  {"details", TDF_DETAILS},
  {"stats", TDF_STATS},
  {"blocks", TDF_BLOCKS},
  {"vops", TDF_VOPS},
  {"lineno", TDF_LINENO},
  {"uid", TDF_UID},
  {"stmtaddr", TDF_STMTADDR},
  {"all", ~(TDF_RAW | TDF_SLIM | TDF_LINENO | TDF_TREE | TDF_RTL | TDF_IPA 
	    | TDF_STMTADDR)},
  {NULL, 0}
};

unsigned int
dump_register (const char *suffix, const char *swtch, const char *glob,
	       int flags, unsigned int num, int letter)
{
  size_t this = extra_dump_files_in_use++;

  if (this >= extra_dump_files_alloced)
    {
      if (extra_dump_files_alloced == 0)
	extra_dump_files_alloced = 32;
      else
	extra_dump_files_alloced *= 2;
      extra_dump_files = xrealloc (extra_dump_files,
				   sizeof (struct dump_file_info)
				   * extra_dump_files_alloced);
    }

  memset (&extra_dump_files[this], 0, sizeof (struct dump_file_info));
  extra_dump_files[this].suffix = suffix;
  extra_dump_files[this].swtch = swtch;
  extra_dump_files[this].glob = glob;
  extra_dump_files[this].flags = flags;
  extra_dump_files[this].num = num;
  extra_dump_files[this].letter = letter;

  return this + TDI_end;
}


/* Return the dump_file_info for the given phase.  */

struct dump_file_info *
get_dump_file_info (enum tree_dump_index phase)
{
  if (phase < TDI_end)
    return &dump_files[phase];
  else if (phase - TDI_end >= extra_dump_files_in_use)
    return NULL;
  else
    return extra_dump_files + (phase - TDI_end);
}


/* Return the name of the dump file for the given phase.
   If the dump is not enabled, returns NULL.  */

char *
get_dump_file_name (enum tree_dump_index phase)
{
  char dump_id[7];
  struct dump_file_info *dfi;

//  if (phase == TDI_none)
//    return NULL;

  dfi = get_dump_file_info (phase);
//  if (dfi->state == 0)
//    return NULL;

//  if (dfi->num < 0)
//    dump_id[0] = '\0';
//a  else
    {
      const char *template;
/*
     if (dfi->flags & TDF_TREE)
	template = ".t%02d";
      else if (dfi->flags & TDF_IPA)
	template = ".i%02d";
      else
*/
	template = ".%02d";

      if (snprintf (dump_id, sizeof (dump_id), template, dfi->num) < 0)
	dump_id[0] = '\0';
    }

  return concat (dump_base_name, dump_id, dfi->suffix, NULL);
}

/* Returns nonzero if tree dump PHASE is enabled.  */

int
dump_enabled_p (enum tree_dump_index phase)
{
 struct dump_file_info *dfi = get_dump_file_info (phase);
 
 if (dfi->state == 0)
 {
     dfi->state =-1;//? mdupont hacked this TODO
 }
  return dfi->state;
}

/* Returns nonzero if tree dump PHASE has been initialized.  */

int
dump_initialized_p (enum tree_dump_index phase)
{
  struct dump_file_info *dfi = get_dump_file_info (phase);
  return dfi->state > 0;
}

/* Returns the switch name of PHASE.  */

const char *
dump_flag_name (enum tree_dump_index phase)
{
  struct dump_file_info *dfi = get_dump_file_info (phase);
  return dfi->swtch;
}


/* Enable all tree dumps.  Return number of enabled tree dumps.  */

int
dump_enable_all (int flags, int letter)
{
  int n = 0;
  size_t i;
  letter=0;
  for (i = TDI_none + 1; i < (size_t) TDI_end; i++)

/*    if ((dump_files[i].flags & flags)
	&& (letter == 0 || letter == dump_files[i].letter))
*/
      {
        dump_files[i].state = -1;
        dump_files[i].flags = flags;
        n++;
      }

  for (i = 0; i < extra_dump_files_in_use; i++)
/*    if ((extra_dump_files[i].flags & flags)
	&& (letter == 0 || letter == extra_dump_files[i].letter))
*/
      {
        extra_dump_files[i].state = -1;
        extra_dump_files[i].flags = flags;
	n++;
      }

  return n;
}

/* Parse ARG as a dump switch. Return nonzero if it is, and store the
   relevant details in the dump_files array.  */

int
dump_switch_p_1 (const char *arg, struct dump_file_info *dfi, bool doglob);

int
dump_switch_p_1 (const char *arg, struct dump_file_info *dfi, bool doglob)
{
  const char *option_value;
  const char *ptr;
  int flags;
  
  if (doglob && !dfi->glob)
    return 0;

  option_value = skip_leading_substring (arg, doglob ? dfi->glob : dfi->swtch);
  if (!option_value)
    return 0;

  ptr = option_value;
  flags = 0;

  while (*ptr)
    {
      const struct dump_option_value_info *option_ptr;
      const char *end_ptr;
      unsigned length;

      while (*ptr == '-')
	ptr++;
      end_ptr = strchr (ptr, '-');
      if (!end_ptr)
	end_ptr = ptr + strlen (ptr);
      length = end_ptr - ptr;

      for (option_ptr = dump_options; option_ptr->name; option_ptr++)
	if (strlen (option_ptr->name) == length
	    && !memcmp (option_ptr->name, ptr, length))
	  {
	    flags |= option_ptr->value;
	    goto found;
	  }
      warning ("ignoring unknown option %q.*s in %<-fdump-%s%>",
	       length, ptr, dfi->swtch);
    found:;
      ptr = end_ptr;
    }

  // introspector
  //TODO : finish up the dumping  intrspctr_finish(di);

  dfi->state = -1;

  flags =~(TDF_RAW | TDF_SLIM | TDF_LINENO | TDF_TREE | TDF_RTL | TDF_IPA); //mdupont hacked

  dfi->flags |= flags;

  /* Process -fdump-tree-all and -fdump-rtl-all, by enabling all the
     known dumps.  */
  if (dfi->suffix == NULL)
    dump_enable_all (dfi->flags, 0);

  return 1;
}

int
dump_switch_p (const char *arg)
{
  size_t i;
  int any = 0;

  for (i = TDI_none + 1; i != TDI_end; i++)
    any |= dump_switch_p_1 (arg, &dump_files[i], false);

  /* Don't glob if we got a hit already */
  if (!any)
    for (i = TDI_none + 1; i != TDI_end; i++)
      any |= dump_switch_p_1 (arg, &dump_files[i], true);

  for (i = 0; i < extra_dump_files_in_use; i++)
    any |= dump_switch_p_1 (arg, &extra_dump_files[i], false);
  
  if (!any)
    for (i = 0; i < extra_dump_files_in_use; i++)
      any |= dump_switch_p_1 (arg, &extra_dump_files[i], true);


  return any;
}

bool
enable_rtl_dump_file (int letter)
{
  if (letter == 'a')
    letter = 0;

  return dump_enable_all (TDF_RTL, letter) > 0;
}


